#include <algorithm>
#include <iostream>
#include <memory>

struct thing {
public:
	thing() { std::cout << "thing constructed\n"; }
	~thing() { std::cout << "thing destroyed\n"; }
	void non_const_m() { std::cout << "non-const method called\n"; }
	void const_m() const { std::cout << "const method called\n"; }
};

// --- these are OK ---

void foo1(std::unique_ptr<thing> ptr) {
	// takes ownership
	// needs to check for null
	std::cout << "in foo1\n";
	if (ptr) {
		ptr->non_const_m();
	} else {
		std::cout << "got null!\n";
	}
	// the memory is deallocated at the end of scope here
}

void foo2(thing* ptr) {
	// does not take ownership
	// needs to check for null
	// may modify the pointed-at object
	std::cout << "in foo2\n";
	if (ptr) {
		ptr->non_const_m();
	} else {
		std::cout << "got null!\n";
	}
}

void foo3(const thing* ptr) {
	// as foo2, but may not modify the pointed-at object
	std::cout << "in foo3\n";
	if (ptr) {
		ptr->const_m();
	} else {
		std::cout << "got null!\n";
	}
}

void foo4(thing& ref) {
	// does not take ownership
	// no need to check for null (it's the responsibility of the caller)
	// may modify the pointed-at object
	std::cout << "in foo4\n";
	ref.non_const_m();
}

void foo5(const thing& ref) {
	// as foo4, but may not modify the referenced object
	std::cout << "in foo5\n";
	ref.const_m();
}

// --- these are problematic ---

void foo6(std::unique_ptr<thing>&& ptr) {
	// from the caller's point of view this looks similar
	// to foo1, but the caller cannot be sure that the ownership
	// has been indeed transferred elsewhere
	std::cout << "in foo6\n";
	auto my_ptr = std::move(ptr);
	if (my_ptr) {
		my_ptr->non_const_m();
	} else {
		std::cout << "got null!\n";
	}
}

void foo7(std::unique_ptr<thing>& ptr) {
	// do not do this, malicious!
	std::cout << "in foo7, before reset\n";
	// foo7 can do anything with ptr, including reset and =
	ptr.reset();
	std::cout << "in foo7, after reset\n";
}

void foo8(const std::unique_ptr<thing>& ptr) {
	// not recommended
	// ve cannot change ptr (via reset or =)
	// but we can change the pointed-at object
	std::cout << "in foo6\n";
	if (ptr) {
		ptr->non_const_m();
	} else {
		std::cout << "got null!\n";
	}
	// use foo2 (or foo4) instead to make your intent clear
}

void foo9(const std::unique_ptr<const thing>& /* ptr */) {
	// we cannot call this with an actual unique_ptr<thing> argument
	// you want to use foo3 (or foo5) instead
}


int main() {
	auto ptr = std::make_unique<thing>();

	// foo1(ptr); // does not work, we cannot copy ownership
	foo1(std::move(ptr)); // we move the ownership
	// ptr does not own anything anymore
	std::cout << "---\n";

	ptr = std::make_unique<thing>();
	foo2(ptr.get());
	foo3(ptr.get());
	std::cout << "---\n";

	// our responsibility that ptr is not null
	foo4(*ptr);
	foo5(*ptr);
	std::cout << "---\n";

	// the bad ones
	foo6(std::move(ptr)); // did we move the ownership or not?
	if (ptr == nullptr) {
		std::cout << "ptr is now null\n";
	}
	std::cout << "---\n";

	ptr = std::make_unique<thing>();
	foo7(ptr); // really bad! we may lose ownership without explicit move
	if (ptr == nullptr) {
		std::cout << "ptr is now null\n";
	}
	std::cout << "---\n";

	ptr = std::make_unique<thing>();
	foo8(ptr);
	std::cout << "---\n";

	// foo9(ptr); // does not work; why?
}
